@@ -28,6 +28,11 @@
 #include <QUrl>
 #include <QProcess>
 #include <sys/utsname.h>
+#include <sys/mman.h>
+#include <sys/stat.h>        /* For mode constants */
+#include <fcntl.h>           /* For O_* constants */
+#include <unistd.h>
+
 #include "stats_dialog.h"
 #include "db.h"
 #include "graph.h"
@@ -57,7 +62,8 @@ static int find_child(int id) {
 	return -1;
 }
 
-StatsDialog::StatsDialog(): QDialog(), mode_(MODE_TOP), pid_(0), uid_(0), lts_(false),
+StatsDialog::StatsDialog(): QDialog(), fdns_report_(0), fdns_seq_(0), fdns_fd_(0), fdns_first_run_(true), fdns_cnt_(0),
+		mode_(MODE_TOP), pid_(0), uid_(0), lts_(false),
 	pid_initialized_(false), pid_seccomp_(false), pid_caps_(QString("")), pid_noroot_(false),
 	pid_cpu_cores_(QString("")), pid_protocol_(QString("")), pid_name_(QString("")),
 	profile_(QString("")), pid_x11_(0),
@@ -116,9 +122,12 @@ StatsDialog::StatsDialog(): QDialog(), mode_(MODE_TOP), pid_(0), uid_(0), lts_(f
 
 	thread_ = new PidThread();
 	connect(thread_, SIGNAL(cycleReady()), this, SLOT(cycleReady()));
+	createTrayActions();
 }
 
 StatsDialog::~StatsDialog() {
+	if (fdns_fd_)
+		::close(fdns_fd_);
 	if (!isMaximized())
 		config_write_screen_size(width(), height());
 }
@@ -129,6 +138,45 @@ void StatsDialog::cleanStorage() {
 	storage_seccomp_ = "";
 	storage_intro_ = "";
 	storage_network_ = "";
+	storage_netfilter_ = "";
+}
+
+// Shutdown sequence
+void StatsDialog::main_quit() {
+	printf("exiting...\n");
+
+	qApp->quit();
+}
+
+void StatsDialog::trayActivated(QSystemTrayIcon::ActivationReason reason) {
+	if (reason == QSystemTrayIcon::Context)
+		return;
+	if (reason == QSystemTrayIcon::DoubleClick)
+		return;
+	if (reason == QSystemTrayIcon::MiddleClick)
+		return;
+
+	if (isVisible()) {
+		hide();
+//		stats_->hide();
+	}
+	else {
+		show();
+//		stats_->hide();
+	}
+}
+
+void StatsDialog::createTrayActions() {
+	minimizeAction = new QAction(tr("Mi&nimize"), this);
+	connect(minimizeAction, SIGNAL(triggered()), this, SLOT(hide()));
+//	connect(minimizeAction, SIGNAL(triggered()), stats_, SLOT(hide()));
+
+	restoreAction = new QAction(tr("&Restore"), this);
+	connect(restoreAction, SIGNAL(triggered()), this, SLOT(showNormal()));
+//	connect(restoreAction, SIGNAL(triggered()), stats_, SLOT(show()));
+
+	quitAction = new QAction(tr("&Quit"), this);
+	connect(quitAction, SIGNAL(triggered()), this, SLOT(main_quit()));
 }
 
 QString StatsDialog::header() {
@@ -138,6 +186,13 @@ QString StatsDialog::header() {
 		msg += "<a href=\"about\">About</a>";
 		msg += "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href=\"launcher\">Sandbox Launcher</a>";
 		msg += "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href=\"newsandbox\">Configuration Wizard</a>";
+		msg += "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;<a href=\"fdns\">Firejail DNS</a>";
+		msg += "</td></tr></table>";
+	}
+
+	else if (mode_ == MODE_FDNS) {
+		msg += "<table><tr><td width=\"5\"></td><td>";
+		msg += "<a href=\"top\">Home</a>";
 		msg += "</td></tr></table>";
 	}
 
@@ -206,6 +261,56 @@ void StatsDialog::updateTop() {
 	procView_->setHtml(msg);
 }
 
+void StatsDialog::updateFdns() {
+	QString msg = header();
+
+	// open fdns shared memory if necessary
+	if (!fdns_fd_) {
+		fdns_fd_ = shm_open("/fdns-stats", O_RDONLY, S_IRWXU);
+		if (fdns_fd_ == -1) {
+			msg += "Error: cannot open /dev/shm/fdns_stats, probably fdns is not running<br/>";
+			fdns_fd_ = 0;
+			procView_->setHtml(msg);
+			return;
+		}
+	}
+
+	if (fdns_fd_ && fdns_report_ == 0) {
+		fdns_report_ = (DnsReport *) mmap(0, sizeof(DnsReport), PROT_READ, MAP_SHARED, fdns_fd_, 0 );
+		if (fdns_report_ == (void *) - 1) {
+			msg += "Error: cannot map /sdv/shm/fdns_stats file in process memory<<br/>";
+			fdns_report_ = 0;
+			::close(fdns_fd_);
+			fdns_fd_ = 0;
+			procView_->setHtml(msg);
+			return;
+		}
+	}
+	if (fdns_fd_ && fdns_report_) {
+		if (fdns_first_run_ || fdns_cnt_++ >= 10 || fdns_seq_ != fdns_report_->seq) {
+			fdns_first_run_ = false;
+			fdns_cnt_ = 0;
+			fdns_seq_ = fdns_report_->seq;
+
+			// print header
+			msg += fdns_report_->header;
+			msg += "<br/>";
+
+			// print log lines
+			for (int i = fdns_report_->logindex; i < MAX_LOG_ENTRIES; i++) {
+				msg += fdns_report_->logentry[i];
+				msg += "<br/>";
+			}
+			for (int i = 0; i < fdns_report_->logindex; i++) {
+				msg += fdns_report_->logentry[i];
+				msg += "<br/>";
+			}
+			procView_->setHtml(msg);
+		}
+	}
+}
+
+
 void StatsDialog::updateFirewall() {
 	DbPid *dbptr = Db::instance().findPid(pid_);
 	if (!dbptr) {
@@ -213,19 +318,21 @@ void StatsDialog::updateFirewall() {
 		return;
 	}
 
-	if (arg_debug)
-		printf("reading firewall configuration\n");
-	QString msg = header() + storage_intro_;
+	QString msg = storage_netfilter_;
+	if (msg.isEmpty()) {
+		if (arg_debug)
+			printf("reading firewall configuration\n");
+		msg = header() + storage_intro_;
 
-	char *cmd;
-	if (asprintf(&cmd, "firejail --netfilter.print=%d", pid_) != -1) {
-		char *str = run_program(cmd);
-		if (str)
-			msg += "<pre>" + QString(str) + "</pre>";
+		char *cmd;
+		if (asprintf(&cmd, "firejail --netfilter.print=%d", pid_) != -1) {
+			char *str = run_program(cmd);
+			if (str)
+				msg += "<pre>" + QString(str) + "</pre>";
+		}
+		storage_netfilter_ = msg;
+		procView_->setHtml(msg);
 	}
-
-	procView_->setHtml(msg);
-
 }
 
 
@@ -858,6 +965,8 @@ void StatsDialog::updatePid() {
 void StatsDialog::cycleReady() {
 	if (mode_ == MODE_TOP)
 		updateTop();
+	else if (mode_ == MODE_FDNS)
+		updateFdns();
 	else if (mode_ == MODE_PID)
 		updatePid();
 	else if (mode_ == MODE_TREE)
@@ -975,6 +1084,8 @@ void StatsDialog::anchorClicked(const QUrl & link) {
 		QMessageBox::about(this, tr("About"), msg);
 
 	}
+	else if (linkstr == "fdns")
+		mode_ = MODE_FDNS;
 	else if (linkstr == "newsandbox") {
 		// start firejail-ui as a separate process
 
@@ -988,7 +1099,7 @@ void StatsDialog::anchorClicked(const QUrl & link) {
 		int rv = system("firetools &");
 		(void) rv;
 	}
-	else {
+	else { // linstr == "home"
 		pid_ = linkstr.toInt();
 		pid_initialized_ = false;
 		pid_caps_ = QString("");
@@ -997,6 +1108,9 @@ void StatsDialog::anchorClicked(const QUrl & link) {
 		mode_ = MODE_PID;
 	}
 
+	// reset fdns
+	fdns_first_run_ = true;
+
 	if (data_ready)
 		cycleReady();
 }
